"""This module contains all the objects that can be used in the game window.
Each object is a subclass of Sprite, which is a subclass of pygame.sprite.Sprite.
The objects are: Box, Circle, Line, Text, and Group.
Each object has a corresponding new_* function that can be used to create the object.
For example, play.new_box() creates a new Box object.
"""

from .box import Box
from .circle import Circle
from .line import Line
from .sprite import Sprite
from .text import Text
from .image import Image
from .sound import Sound



"""All the events that can be triggered in the game."""

from ..callback import callback_manager, CallbackType
from ..io.keypress import (
    when_key as _when_key,
    when_any_key as _when_any_key,
)
from ..io.controllers import controllers
from ..io.mouse import mouse
from ..utils.async_helpers import _make_async
from ..callback.callback_helpers import run_async_callback

_ = controllers  # Work around for the global variable not being imported


# @decorator
def when_program_starts(func):
    """
    Call code right when the program starts.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_program_starts
        def show():
            text.words = 'started'

    :param func: The function to call when the program starts.
    :return: The decorator function.
    """
    async_callback = _make_async(func)

    async def wrapper():
        return await run_async_callback(
            async_callback,
            [],
            [],
        )

    callback_manager.add_callback(CallbackType.WHEN_PROGRAM_START, wrapper)
    return func


def repeat(number_of_times):
    """
    Repeat a set of commands a certain number of times.
    Equivalent to `range(1, number_of_times+1)`.
    :param number_of_times: The number of times to repeat the commands.
    :return: A range object that can be iterated over.
    """
    return range(1, number_of_times + 1)


# @decorator
def repeat_forever(func):
    """
    Calls the given function repeatedly in the game loop.

    Example:

        text = play.new_text(words='hi there!')

        @play.repeat_forever
        def turn():
            print("this is printing every single frame")


    :param func: The function to call repeatedly.
    :return: The decorator function.
    """
    async_callback = _make_async(func)

    async def repeat_wrapper():
        repeat_wrapper.is_running = True
        await run_async_callback(
            async_callback,
            [],
            [],
        )
        repeat_wrapper.is_running = False

    repeat_wrapper.is_running = False
    callback_manager.add_callback(CallbackType.REPEAT_FOREVER, repeat_wrapper)
    return func


# @decorator
def when_sprite_clicked(*sprites):
    """A decorator that runs a function when a sprite is clicked.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_sprite_clicked(text)
        def show(sprite):
            text.words = 'clicked!'

    :param sprites: The sprites to run the function on.
    :return: The function to run.
    """

    def wrapper(func):
        for sprite in sprites:
            sprite.when_clicked(func, call_with_sprite=True)
        return func

    return wrapper


# @decorator
def when_any_key_pressed(func):
    """
    Calls the given function when any key is pressed.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_any_key_pressed
        def show(key):
            text.words = key
    """
    if not callable(func):
        raise ValueError("""@play.when_any_key_pressed doesn't use a list of keys.""")
    return _when_any_key(func, released=False)


# @decorator
def when_key_pressed(*keys):
    """
    Calls the given function when any of the specified keys are pressed.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_key_pressed('a')
        def show(key):
            text.words = key
    """
    return _when_key(*keys, released=False)


# @decorator
def when_any_key_released(func):
    """
    Calls the given function when any key is released.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_any_key_released
        def show(key):
            text.words = key
    """
    if not callable(func):
        raise ValueError("""@play.when_any_key_released doesn't use a list of keys.""")
    return _when_any_key(func, released=True)


# @decorator
def when_key_released(*keys):
    """
    Calls the given function when any of the specified keys are released.

    Example:

        text = play.new_text(words='hi there!')

        @play.when_key_released('a')
        def show(key):
            text.words = key
    """
    return _when_key(*keys, released=True)


# @decorator
def when_mouse_clicked(func):
    """
    Calls the given function when the mouse is clicked.

    Example:
        text = play.new_text(words='hi there!')

        @play.when_mouse_clicked
        def show():
            text.words = 'mouse click'
    """
    return mouse.when_clicked(func)


# @decorator
def when_click_released(func):
    """
    Calls the given function when the mouse click is released.

    Example:
        text = play.new_text(words='hi there!')

        @play.when_click_released
        def show():
            text.words = 'mouse click released'
    """
    return mouse.when_click_released(func)



"""Handles custom music/sound being played."""

import pygame
from ..io.logging import play_logger as logger


class Sound:
    def __init__(self, file_name, volume=1.0, loops=1):
        """
        Initialize the Sound object.
        :param file_name: The sound file to load.
        :param volume: The initial volume (0.0 to 1.0).
        :param loops: Number of times to loop the sound (-1 for infinite).
        """
        pygame.mixer.init()
        self.sound = None
        self.channel = None
        self.file_path = file_name
        self.volume = volume
        self.loops = loops
        self.is_paused = False
        self.load(file_name)
        self.set_volume(volume)

    def load(self, file_name):
        """Load a sound file."""
        try:
            self.sound = pygame.mixer.Sound(file_name)

        except FileNotFoundError:
            logger.error("File not found", exc_info=True)
        logger.info("Loaded sound file", exc_info=True)

    def play(self):
        """Play the loaded sound with the specified loop settings, or resume a paused sound."""
        if not self.sound:
            logger.warning(
                "No sound loaded. Use the 'load' method first.", exc_info=True
            )

        self.channel = pygame.mixer.find_channel()
        if self.channel is None:
            logger.warning("No available channels to play the sound.", exc_info=True)

        if not self.is_playing():
            self.channel.play(self.sound, loops=self.loops)
        if self.is_paused:
            self.channel.unpause()
            self.is_paused = False

    def pause(self):
        """Pause the sound."""
        if self.channel.get_busy():
            self.channel.pause()
            self.is_paused = True

    def length_song(self):
        """Returns the length of the song as a float"""

        return round((self.channel.get_sound().get_length()), 2)

    def stop(self):
        """Stop current channel"""
        self.channel.stop()

    def set_volume(self, volume):
        """Set the volume of the sound (0.0 to 1.0)."""
        if not self.sound:
            logger.warning(
                "No sound loaded. Use the 'load' method first.", exc_info=True
            )
        if not 0.0 <= volume <= 1.0:
            logger.warning("Volume must be between 0.0 and 1.0", exc_info=True)
        self.volume = volume
        self.sound.set_volume(volume)

    def get_volume(self):
        """Get the current volume of the sound."""
        if not self.sound:
            logger.warning(
                "No sound loaded. Use the 'load' method first.", exc_info=True
            )
        volume = self.sound.get_volume()
        return volume

    def is_playing(self):
        """Check if the sound is currently playing."""
        if self.channel:
            return self.channel.get_busy()
        return False



"""
This module contains the CallbackManager class and CallbackType enum.
"""

from enum import Enum


class CallbackType(Enum):
    REPEAT_FOREVER = 0
    WHEN_PROGRAM_START = 1
    PRESSED_KEYS = 2
    RELEASED_KEYS = 3
    WHEN_CLICKED = 4
    WHEN_CLICK_RELEASED = 5
    WHEN_CLICKED_SPRITE = 6
    WHEN_TOUCHING = 7
    WHEN_CONTROLLER_BUTTON_PRESSED = 8
    WHEN_CONTROLLER_BUTTON_RELEASED = 9
    WHEN_CONTROLLER_AXIS_MOVED = 10


class CallbackManager:
    def __init__(self):
        """
        A class to manage callbacks.
        """
        self.callbacks = {}

    def add_callback(
        self, callback_type, callback, callback_discriminator=None
    ) -> None:
        """
        Add a callback to the callback manager.
        :param callback_type: The type of callback.
        :param callback: The callback function.
        :param callback_discriminator: The discriminator for the callback.
        :return: None
        """
        if callback_type not in self.callbacks:
            if callback_discriminator is None:
                self.callbacks[callback_type] = []
            else:
                self.callbacks[callback_type] = {}
        if callback_discriminator is None:
            self.callbacks[callback_type].append(callback)
        else:
            if callback_discriminator not in self.callbacks[callback_type]:
                self.callbacks[callback_type][callback_discriminator] = []
            self.callbacks[callback_type][callback_discriminator].append(callback)

    def get_callbacks(self, callback_type) -> dict:
        """
        Get the callbacks of a certain type.
        :param callback_type: The type of callback.
        :return: The callbacks of the specified type.
        """
        return self.callbacks.get(callback_type, None)

    def get_callback(self, callback_type, callback_discriminator=None) -> callable:
        """
        Get a callback of a certain type.
        :param callback_type: The type of callback.
        :param callback_discriminator: The discriminator for the callback.
        :return: The callback(s) of the specified type.
        """
        if callback_discriminator is None:
            return self.callbacks.get(callback_type, None)
        return self.callbacks.get(callback_type, {}).get(callback_discriminator, None)


callback_manager = CallbackManager()



"""Generators for creating new objects."""

from ..objects import (
    Box as _Box,
    Circle as _Circle,
    Line as _Line,
    Text as _Text,
    Image as _Image,
    Sound as _Sound,
)


def new_text(  # pylint: disable=too-many-arguments
    words: str = "",
    x: int = 0,
    y: int = 0,
    font: str = "/path/to/font",
    font_size: int = 50,
    color: str = "black",
    angle: int = 0,
    transparency: int = 100,
    size: int = 100,
) -> _Text:
    """Make a new text object.
    :param words: The text to display.
    :param x: The x-coordinate of the text.
    :param y: The y-coordinate of the text.
    :param font: The font to use.
    :param font_size: The size of the font.
    :param color: The color of the text.
    :param angle: The angle of the text.
    :param transparency: The transparency of the text.
    :param size: The size of the text.
    :return: A new text object.
    """
    if not isinstance(words, str):
        raise TypeError("words for a text object must be a string")

    return _Text(
        words=words,
        x=x,
        y=y,
        font=font,
        font_size=font_size,
        color=color,
        angle=angle,
        transparency=transparency,
        size=size,
    )


def new_box(  # pylint: disable=too-many-arguments
    color: str = "black",
    x: int = 0,
    y: int = 0,
    width: int = 100,
    height: int = 200,
    border_color: str = "light blue",
    border_width: int = 0,
    border_radius: int = 0,
    angle: int = 0,
    transparency: int = 100,
    size: int = 100,
) -> _Box:
    """Make a new box object.
    :param color: The color of the box.
    :param x: The x-coordinate of the box.
    :param y: The y-coordinate of the box.
    :param width: The width of the box.
    :param height: The height of the box.
    :param border_color: The color of the border of the box.
    :param border_width: The width of the border of the box.
    :param border_radius: The radius of the border (rounding).
    :param angle: The angle of the box.
    :param transparency: The transparency of the box.
    :param size: The size of the box.
    :return: A new box object.
    """
    return _Box(
        color=color,
        x=x,
        y=y,
        width=width,
        height=height,
        border_color=border_color,
        border_width=border_width,
        border_radius=border_radius,
        angle=angle,
        transparency=transparency,
        size=size,
    )


def new_circle(  # pylint: disable=too-many-arguments
    color: str = "black",
    x: int = 0,
    y: int = 0,
    radius: int = 100,
    border_color: str = "light blue",
    border_width: int = 0,
    transparency: int = 100,
    size: int = 100,
    angle: int = 0,
) -> _Circle:
    """Make a new circle object.
    :param color: The color of the circle.
    :param x: The x-coordinate of the circle.
    :param y: The y-coordinate of the circle.
    :param radius: The radius of the circle.
    :param border_color: The color of the border of the circle.
    :param border_width: The width of the border of the circle.
    :param transparency: The transparency of the circle.
    :param size: The size of the circle.
    :param angle: The angle of the circle.
    :return: A new circle object.
    """
    return _Circle(
        color=color,
        x=x,
        y=y,
        radius=radius,
        border_color=border_color,
        border_width=border_width,
        transparency=transparency,
        size=size,
        angle=angle,
    )


def new_line(  # pylint: disable=too-many-arguments
    color: str = "black",
    x: int = 0,
    y: int = 0,
    length=None,
    angle=None,
    thickness: int = 1,
    x1=None,
    y1=None,
    transparency: int = 100,
    size: int = 100,
) -> _Line:
    """Make a new line object.
    :param color: The color of the line.
    :param x: The x-coordinate of the line.
    :param y: The y-coordinate of the line.
    :param length: The length of the line.
    :param angle: The angle of the line.
    :param thickness: The thickness of the line.
    :param x1: The x-coordinate of the second point of the line.
    :param y1: The y-coordinate of the second point of the line.
    :param transparency: The transparency of the line.
    :param size: The size of the line.
    :return: A new line object.
    """
    return _Line(
        color=color,
        x=x,
        y=y,
        length=length,
        angle=angle,
        thickness=thickness,
        x1=x1,
        y1=y1,
        transparency=transparency,
        size=size,
    )


def new_image(  # pylint: disable=too-many-arguments
    image: str = "/path/to/image",
    x: int = 0,
    y: int = 0,
    size: int = 100,
    angle: int = 0,
    transparency: int = 100,
) -> _Image:
    """Make a new image object.
    :param image: The image to display.
    :param x: The x-coordinate of the image.
    :param y: The y-coordinate of the image.
    :param size: The size of the image.
    :param angle: The angle of the image.
    :param transparency: The transparency of the image.
    :return: A new image object.
    """
    return _Image(
        image=image, x=x, y=y, size=size, angle=angle, transparency=transparency
    )


def new_sound(  # pylint: disable=too-many-arguments
    file_name: str = "file.mp3",
    volume: float = 1.0,
    loops: int = 0,
) -> _Sound:
    """
    Initialize the Sound object.
    :param file_name: The sound file to load (file path if not in the same directory as the .py).
    :param volume: The initial volume (0.0 to 1.0).
    :param loops: Number of times to loop the sound (-1 for infinite, 0 for no loop).
    """

    return _Sound(file_name=file_name, volume=volume, loops=loops)



"""This module contains the controller loop, which handles controller events in the game loop."""

import pygame  # pylint: disable=import-error

from ..callback import callback_manager, CallbackType
from ..io.controllers import (
    controllers,
)
from ..callback.callback_helpers import run_callback

controller_axis_moved = False  # pylint: disable=invalid-name
controller_button_pressed = False  # pylint: disable=invalid-name
controller_button_released = False  # pylint: disable=invalid-name


def _handle_controller_events(event):
    """Handle controller events in the game loop.
    :param event: The event to handle."""
    if event.type == pygame.JOYAXISMOTION:  # pylint: disable=no-member
        global controller_axis_moved
        controller_axis_moved = True
    if event.type == pygame.JOYBUTTONDOWN:  # pylint: disable=no-member
        global controller_button_pressed
        controller_button_pressed = True
    if event.type == pygame.JOYBUTTONUP:
        global controller_button_released
        controller_button_released = True


def _handle_controller():  # pylint: disable=too-many-branches
    """Handle controller events in the game loop."""
    ############################################################
    # @controller.when_button_pressed and @controller.when_any_button_pressed
    ############################################################
    global controller_button_pressed, controller_button_released, controller_axis_moved
    if controller_button_pressed and callback_manager.get_callbacks(
        CallbackType.WHEN_CONTROLLER_BUTTON_PRESSED
    ):
        controller_button_callbacks = callback_manager.get_callbacks(
            CallbackType.WHEN_CONTROLLER_BUTTON_PRESSED
        )
        if "any" in controller_button_callbacks:
            for callback in controller_button_callbacks["any"]:
                for button in range(controllers.get_numbuttons(callback.controller)):
                    if (
                        controllers.get_controller(callback.controller).get_button(
                            button
                        )
                        == 1
                    ):
                        run_callback(callback, ["button_number"], [], button)
        for button, callbacks in controller_button_callbacks.items():
            if button != "any":
                for callback in callbacks:
                    if controllers.get_button(callback.controller, button) == 1:
                        run_callback(callback, ["button_number"], [], [], button)
        controller_button_pressed = False
    ############################################################
    # @controller.when_button_released
    ############################################################
    if controller_button_released and callback_manager.get_callbacks(
        CallbackType.WHEN_CONTROLLER_BUTTON_RELEASED
    ):
        released_callbacks = callback_manager.get_callbacks(
            CallbackType.WHEN_CONTROLLER_BUTTON_RELEASED
        )
        if "any" in released_callbacks:
            for callback in released_callbacks["any"]:
                for button in range(controllers.get_numbuttons(callback.controller)):
                    if (
                        controllers.get_controller(callback.controller).get_button(
                            button
                        )
                        == 0
                    ):
                        run_callback(callback, ["button_number"], [], button)
        for button, callbacks in released_callbacks.items():
            for callback in callbacks:
                if controllers.get_button(callback.controller, button) == 0:
                    run_callback(callback, ["button_number"], [], button)
        controller_button_released = False
    ############################################################
    # @controller.when_axis_moved
    ############################################################
    if controller_axis_moved and callback_manager.get_callbacks(
        CallbackType.WHEN_CONTROLLER_AXIS_MOVED
    ):
        axis_moved_callbacks = callback_manager.get_callbacks(
            CallbackType.WHEN_CONTROLLER_AXIS_MOVED
        )
        if "any" in axis_moved_callbacks:
            for callback in axis_moved_callbacks["any"]:
                for axis in range(controllers.get_numaxes(callback.controller)):
                    run_callback(
                        callback,
                        ["axis_number", "axis_value"],
                        [],
                        axis,
                        controllers.get_axis(callback.controller, axis),
                    )
        for axis, callbacks in axis_moved_callbacks.items():
            if axis != "any":
                for callback in callbacks:
                    run_callback(
                        callback,
                        ["axis_number", "axis_value"],
                        [],
                        axis,
                        controllers.get_axis(callback.controller, axis),
                    )
        controller_axis_moved = False



"""
This module contains a decorator that listens to exceptions in the game loop.
"""

from ..io.logging import play_logger
from ..loop import loop as _loop


# @decorator
def listen_to_failure():
    """
    A decorator that listens to exceptions in the game loop.
    """

    def decorate(f):
        def applicator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                _loop.stop()
                play_logger.critical(  # pylint: disable=logging-fstring-interpolation
                    f"Error in {f.__name__}: {e}"
                )
                raise e

        return applicator

    return decorate



"""Game functions and utilities."""

import asyncio as _asyncio
import logging as _logging

import pygame  # pylint: disable=import-error

from ..callback import callback_manager, CallbackType
from ..core import game_loop as _game_loop
from ..callback.callback_helpers import run_callback
from ..loop import loop as _loop
from ..utils import color_name_to_rgb as _color_name_to_rgb
from ..io.keypress import _pressed_keys
from ..globals import backdrop as __backdrop
from ..physics import set_physics_simulation_steps as _set_physics_simulation_steps

_backdrop = __backdrop  # Work around for the global variable not being imported


def start_program():
    """
    Calling this function starts your program running.

    play.start_program() should almost certainly go at the very end of your program.
    """
    if callback_manager.get_callbacks(CallbackType.WHEN_PROGRAM_START) is not None:
        for func in callback_manager.get_callbacks(CallbackType.WHEN_PROGRAM_START):
            run_callback(
                func,
                [],
                [],
            )

    _loop.call_soon(_game_loop)
    try:
        _loop.run_forever()
    finally:
        logger = _logging.getLogger("asyncio")
        logger.setLevel(_logging.CRITICAL)
        pygame.quit()  # pylint: disable=no-member


def stop_program():
    """
    Calling this function stops your program running.

    play.stop_program() should almost certainly go at the very end of your program.
    """
    _loop.stop()
    pygame.quit()  # pylint: disable=no-member


async def animate():
    """
    Wait for the next frame to be drawn.
    """
    await _asyncio.sleep(0)


def set_backdrop(color_or_image_name):
    """Set the backdrop color or image for the game.
    :param color_or_image_name: The color or image to set as the backdrop.
    """
    global _backdrop
    _color_name_to_rgb(color_or_image_name)

    _backdrop = color_or_image_name


async def timer(seconds=1.0):
    """Wait a number of seconds. Used with the await keyword like this:
    :param seconds: The number of seconds to wait.
    :return: True after the number of seconds has passed.
    """
    await _asyncio.sleep(seconds)
    return True


def key_is_pressed(*keys):
    """
    Returns True if any of the given keys are pressed.

    Example:

        @play.repeat_forever
        async def do():
            if play.key_is_pressed('up', 'w'):
                print('up or w pressed')
    """

    for key in keys:
        if key in _pressed_keys.values():
            return True
    return False


def set_physics_simulation_steps(num_steps: int) -> None:
    """
    Set the number of simulation steps for the physics engine.
    :param num_steps: The number of simulation steps.
    """
    _set_physics_simulation_steps(num_steps)
